import re
from copy import copy
from typing import Dict, List, Literal

import yarl
from httpx import Client
from pyquery import PyQuery as pq


BASE_URL = "https://b1.facecast.io/"
BASE_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:28.0) Gecko/20100101 Firefox/28.0",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "en-GB,en-US;q=0.9,en;q=0.8",
}
AJAX_HEADERS = copy(BASE_HEADERS)
AJAX_HEADERS.update(
    {
        "Accept": "application/json, text/javascript, */*; q=0.01",
        "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
        "x-requested-with": "XMLHttpRequest",
    }
)


class AuthError(Exception):
    ...


class ServerConnector:
    def __init__(self, client: Client):
        self.client = client
        self.is_authorized: bool = False
        self.form_sign = None

    def _update_from_sign(self):
        r = self.client.get("en/main")
        match = re.search(r"form_sign: \'(\w+)\'", r.text)
        if match:
            self.form_sign = match.group(1)
            return
        raise AuthError("Failed to fetch form_sign")

    def _fetch_signature(self, text):
        match = re.search(r"signature: \'(\w+)\'", text)
        if match:
            return match.group(1)
        raise AuthError("Failed to fetch signature")

    def do_auth(self, username: str, password: str) -> bool:
        r = self.client.get("en/main")
        if r.url == "en/main":
            self.is_authorized = True
            return True
        signature = self._fetch_signature(r.text)
        r = self.client.post(
            "en/login",
            params={"mode": "ajaj"},
            data={"login": username, "pass": password, "signature": signature,},
            headers=AJAX_HEADERS,
        )
        if r.status_code == 200 and r.json().get("ok"):
            self.is_authorized = True
            self._update_from_sign()
            return True
        self.is_authorized = False
        raise AuthError("AuthService error")

    def get_devices(self) -> List[Dict]:
        r = self.client.get("en/main")
        d = pq(r.text)
        devices = d(".sb-streamboxes-main-list a")
        devices_names = devices.find(".sb-streambox-item-name")
        return [
            dict(rtmp_id=device.attrib["href"].split("=")[1], name=device_name.text,)
            for device, device_name in zip(devices, devices_names)
        ]

    def get_device(self, rtmp_id: str) -> Dict:
        r = self.client.post(
            "en/rtmp",
            data={"action": "get_info"},
            params={"rtmp_id": rtmp_id},
            headers=AJAX_HEADERS,
        )
        return r.json()

    def create_device(self, name: str) -> bool:
        r = self.client.post(
            "en/main_add/ajaj",
            data={"cmd": "add_rtmp", "sbin": 0, "sign": self.form_sign, "title": name},
        )
        return r.status_code == 200

    def delete_device(self, rtmp_id: str) -> bool:
        r = self.client.post(
            "en/rtmp_popup_menu/ajaj",
            data={
                "cmd": "delete_rtmp_source",
                "sign": self.form_sign,
                "rtmp_id": rtmp_id,
            },
        )
        return r.status_code == 200

    def get_input_params(self, rtmp_id: str) -> Dict:
        r = self.client.get("en/rtmp", params={"rtmp_id": rtmp_id})
        d = pq(r.text)
        server_url = d(".sb-input-input-url").attr["value"]
        if yarl.URL(server_url).host is None:
            return self.get_input_params(rtmp_id)
        shared_key = d(".sb-input-sharedkey").attr["value"]
        return dict(server_url=server_url, shared_key=shared_key)

    def get_status(self, rtmp_id: str) -> Dict:
        r = self.client.post(
            "en/rtmp/ajaj",
            data={
                "sign": self.form_sign,
                "requests[0][cmd]": "get_status",
                "requests[0][sign]": self.form_sign,
                "requests[1][cmd]": "input_status",
                "requests[1][sign]": self.form_sign,
                "requests[2][cmd]": "output_status",
                "requests[2][sign]": self.form_sign,
                "rtmp_id": rtmp_id,
            },
            params={"rtmp_id": rtmp_id},
            headers=AJAX_HEADERS,
        )
        return r.json()

    def get_outputs(self, rtmp_id: str) -> Dict:
        r = self.client.post(
            "en/rtmp_outputs/ajaj",
            data={"cmd": "getlist", "rtmp_id": rtmp_id, "sign": self.form_sign},
            params={"rtmp_id": rtmp_id},
            headers=AJAX_HEADERS,
        )
        return r.json()

    def update_output(
        self,
        rtmp_id: str,
        output_id: str,
        server_url: str,
        shared_key: str,
        title: str,
        audio: int = 0,
    ):
        r = self.client.post(
            "en/out_rtmp_rtmp/ajaj",
            data={
                "cmd": "update",
                "rtmp_id": rtmp_id,
                "oid": output_id,
                "sign": self.form_sign,
                "server_url": server_url,
                "shared_key": shared_key,
                "title": title,
                "audio": audio,
            },
            headers=AJAX_HEADERS,
        )
        return r.json()

    def create_output(
        self,
        rtmp_id: str,
        server_url: str,
        shared_key: str,
        title: str,
        audio: int = 0,
        stream_type: Literal["rtmp", "mpegts"] = "rtmp",
    ) -> Dict:
        r = self.client.post(
            "en/out_rtmp_rtmp/ajaj",
            data={
                "cmd": "add",
                "rtmp_id": rtmp_id,
                "oid": 0,
                "sign": self.form_sign,
                "server_url": server_url,
                "shared_key": shared_key,
                "descr": title,
                "type": stream_type,
                "audio": audio,
            },
            headers=AJAX_HEADERS,
        )
        return r.json()

    def delete_output(self, rtmp_id: str, oid: str) -> Dict:
        r = self.client.post(
            "en/out_rtmp_rtmp/ajaj",
            data={
                "cmd": "delete",
                "sign": self.form_sign,
                "oid": oid,
                "rtmp_id": rtmp_id,
            },
            headers=AJAX_HEADERS,
        )
        return r.json()

    def _output_management(self, oid, cmd):
        r = self.client.post(
            "en/out_rtmp_rtmp/ajaj",
            data={
                "cmd": cmd,
                "sign": self.form_sign,
                "oid": oid,
                "ignore_dirty_buffer": False,
            },
            headers=AJAX_HEADERS,
        )
        return r.json()

    def start_output(self, oid: str) -> Dict:
        return self._output_management(oid, "start")

    def stop_output(self, oid: str) -> Dict:
        return self._output_management(oid, "stop")

    def select_server(self, rtmp_id, server_id):
        r = self.client.post(
            "en/rtmp",
            data={
                "cmd": "select_server",
                "sign": self.form_sign,
                "rtmp_id": rtmp_id,
                "server_id": server_id,
            },
            params={"rtmp_id": rtmp_id},
            headers=AJAX_HEADERS,
        )
        return r.json()



from __future__ import annotations

import logging
from typing import List, Optional

import attr
from attr import dataclass

from .server_connector import ServerConnector
from .utils import to_bool


@dataclass
class DeviceInputStatus:
    input: DeviceInput
    ok: bool
    msg: str
    response: bool
    status: str
    resolution: str
    fps: str
    time: float


@dataclass
class DeviceStatus:
    device: Device
    ok: bool
    ping: int
    server: str
    server_id: str
    connected: str
    input_url: str
    sharedkey: str
    time: float
    is_online: int


@dataclass
class DeviceInput:
    device: Device

    server_url: str
    shared_key: str
    status: DeviceInputStatus = None

    def update(self):
        input_status = self.device.server_connector.get_status(self.device.rtmp_id).get(
            "input_status"
        )
        if input_status:
            self.status = DeviceInputStatus(
                input=self,
                ok=to_bool(input_status.get("ok")),
                msg=input_status.get("msg"),
                response=input_status.get("response"),
                resolution=input_status.get("resolution"),
                status=input_status.get("status"),
                fps=input_status.get("fps"),
                time=input_status.get("time"),
            )


@dataclass
class DeviceOutputStatus:
    output: DeviceOutput


@dataclass
class DeviceOutput:
    device: Device
    title: str
    enabled: bool
    type: str
    id: str
    cloud: int
    status: DeviceOutputStatus = None

    def update(self):
        output_status = self.device.server_connector.get_status(
            self.device.rtmp_id
        ).get("output_")
        if output_status:
            self.status = DeviceOutputStatus(output=self,)

    def start(self):
        return self.device.server_connector.start_output(self.id)

    def stop(self):
        return self.device.server_connector.stop_output(self.id)

    def delete(self):
        return self.device.server_connector.delete_output(self.device.rtmp_id, self.id)


@dataclass
class Device:
    server_connector: ServerConnector
    name: str
    rtmp_id: str
    online: bool = None
    type: str = None
    lang: str = None
    updates: bool = None
    outputs: List[DeviceOutput] = attr.attrib(factory=list)
    input: DeviceInput = None
    status: DeviceStatus = None

    def _update_device_info(self):
        device_data = self.server_connector.get_device(self.rtmp_id)
        self.online = device_data["online"]
        self.type = device_data["type"]
        self.lang = device_data["lang"]
        self.updates = device_data["updates"]

    def _update_device_status(self):
        status = self.server_connector.get_status(self.rtmp_id).get("get_status")
        if status:
            self.status = DeviceStatus(
                device=self,
                ok=to_bool(status.get("ok")),
                ping=status.get("ping"),
                server=status.get("server"),
                server_id=status.get("server_id"),
                connected=status.get("connected"),
                input_url=status.get("input_url"),
                sharedkey=status.get("sharedkey"),
                time=status.get("time"),
                is_online=status.get("is_online"),
            )

    def _update_outputs(self):
        device_outputs = self.server_connector.get_outputs(self.rtmp_id)
        self.outputs.clear()
        for o in device_outputs:
            do = DeviceOutput(
                device=self,
                title=o["descr"],
                enabled=bool(o["enabled"]),
                type=o["type"],
                id=o["id"],
                cloud=o["cloud"],
            )
            do.update()
            self.outputs.append(do)

    def _update_input(self):
        device_input = self.server_connector.get_input_params(self.rtmp_id)
        di = DeviceInput(
            device=self,
            server_url=device_input["server_url"],
            shared_key=device_input["shared_key"],
        )
        di.update()
        self.input = di

    def update(self):
        self._update_device_info()
        self._update_device_status()
        self._update_input()
        self._update_outputs()

    def create_output(
        self, name, server_url, shared_key, audio=0
    ) -> Optional[DeviceOutput]:
        r = self.server_connector.create_output(
            self.rtmp_id,
            server_url=server_url,
            shared_key=shared_key,
            title=name,
            audio=audio,
        )
        if not to_bool(r["ok"]):
            logging.error(r["msg"])
            return
        o = r["outputs"][0]

        do = DeviceOutput(
            device=self,
            title=o["descr"],
            enabled=bool(o["enabled"]),
            type=o["type"],
            id=o["id"],
            cloud=o["cloud"],
        )
        do.update()
        self.outputs.append(do)
        return do

    def start_outputs(self):
        for o in self.outputs:
            o.start()

    def stop_outputs(self):
        for o in self.outputs:
            o.stop()

    def delete_outputs(self):
        for o in self.outputs:
            o.delete()



from __future__ import absolute_import

import functools
import logging
from time import sleep
from typing import List, Optional, Literal, Dict, TypedDict, Tuple

import httpx

from .models import Device
from facecast_io.server_connector import (
    ServerConnector,
    BASE_URL,
    BASE_HEADERS,
)


class APIError(Exception):
    ...


def auth_required(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.is_authorized:
            raise APIError("Need to authorize first")
        return func(self, *args, **kwargs)

    return wrapper


class Stream(TypedDict):
    name: str
    server_url: str
    shared_key: str


SERVER_LIST = Literal[
    19,
]


class FacecastAPI:
    def __init__(self, auth: Optional[tuple] = None):
        self.client = httpx.Client(
            base_url=BASE_URL,
            verify=False,
            headers=BASE_HEADERS,
            # proxies={
            #     "http": "http://localhost:8000",
            #     "https": "https://localhost:8000",
            # },
        )
        self.server_connector = ServerConnector(self.client)
        if auth:
            self.do_auth(auth[0], auth[1])

    @property
    def is_authorized(self):
        return self.server_connector.is_authorized

    def do_auth(self, username, password):
        return self.server_connector.do_auth(username, password)

    @auth_required
    def get_devices(self, *, update=False) -> List[Device]:
        devices = []
        for d in self.server_connector.get_devices():
            device = Device(
                server_connector=self.server_connector,
                name=d["name"],
                rtmp_id=d["rtmp_id"],
            )
            if update:
                device.update()
            devices.append(device)
        return devices

    @auth_required
    def delete_device(self, name):
        dev = self.get_device(name)
        outputs = self.server_connector.get_outputs(dev.rtmp_id)
        for o in outputs:
            self.server_connector.delete_output(o["id"])
        self.server_connector.delete_device(dev.rtmp_id)

    @auth_required
    def create_new_device(self, name: str, server_id: SERVER_LIST) -> Device:
        if self.server_connector.create_device(name):
            sleep(5)
            device = self.get_device(name, update=True)
            result = self.server_connector.select_server(device.rtmp_id, server_id)
            device.input.shared_key = result["sharedkey"]
            return device
        raise APIError("Some error happened during creation")

    @auth_required
    def get_or_create_device(self, name: str, server_id: SERVER_LIST) -> Device:
        device = self.get_device(name, update=True)
        if device:
            return device
        return self.create_new_device(name, server_id)

    @auth_required
    def create_device_and_outputs(
        self, name, server_id: SERVER_LIST, streams_data: List[Stream]
    ) -> Device:
        device = self.get_or_create_device(name, server_id)
        device.delete_outputs()
        for stream in streams_data:
            output = device.create_output(
                name=stream["name"],
                server_url=stream["server_url"],
                shared_key=stream["shared_key"],
            )
            logging.info(f"{device.name} {output}")
            sleep(3)
        return device

    @auth_required
    def get_device(self, name, update=False) -> Optional[Device]:
        devices = [d for d in self.get_devices() if d.name == name]
        if devices:
            device = devices[-1]
            if update:
                device.update()
            return device

    @auth_required
    def start_outputs(self):
        for d in self.get_devices():
            d.start_outputs()
            logging.info(f"Started for device {d.name}")

    @auth_required
    def stop_outputs(self):
        for d in self.get_devices():
            d.stop_outputs()
            logging.info(f"Stopped for device {d.name}")

    def get_devices_input(self):
        return [
            Stream(
                name=d.name,
                server_url=d.input.server_url,
                shared_key=d.input.shared_key,
            )
            for d in self.get_devices(update=True)
        ]



def to_bool(bool_str):
    """Parse the string and return the boolean value encoded or raise an exception"""
    if isinstance(bool_str, str) and bool_str:
        if bool_str.lower() in ["true", "t", "1"]:
            return True
        elif bool_str.lower() in ["false", "f", "0"]:
            return False
    if isinstance(bool_str, int):
        return bool(bool_str)
    # if here we couldn't parse it
    raise ValueError("%s is no recognized as a boolean value" % bool_str)
